package session

import (
	"context"
	crand "crypto/rand"
	"crypto/sha256"
	"database/sql"
	"errors"
	"fmt"
	"html/template"
	"io/fs"
	"math/rand"
	"net/http"
	"sync"
	"time"

	"github.com/jmoiron/sqlx"
	log "github.com/sirupsen/logrus"

	"git.c3pb.de/gbe/invinoveritas/auth"
)

func hashPassword(ctx context.Context, db *sqlx.DB, user, pass string) (string, error) {
	// Look up password salt from DB and hash password with it
	var salt []byte

	err := db.GetContext(ctx, &salt, `SELECT val FROM state WHERE key = "pwsalt"`)
	if err != nil {
		return "", fmt.Errorf("loading password salt: %w", err)
	}

	h := sha256.New()

	_, err = h.Write(salt)
	if err != nil {
		return "", err
	}

	fmt.Fprint(h, ":", user, ":", pass)

	return fmt.Sprintf("%02x", h.Sum(nil)), nil
}

type Info struct {
	Name    string    `db:"name"`
	Created time.Time `db:"created"`
	Remote  string    `db:"remote"`
}

type Provider struct {
	DB *sqlx.DB
}

// Valid looks up the provided session token in a's database and returns the appropriate user if the token is
// valid. If it is not, an error is returned.
func (a Provider) Valid(ctx context.Context, token string) (*auth.User, error) {
	var user auth.User

	err := a.DB.GetContext(ctx, &user.Name, `
		SELECT users.name
		FROM users, sessions
		WHERE sessions.user == id(users) && sessions.token = ?1`, token)
	if errors.Is(err, sql.ErrNoRows) {
		// Let's see if there are any users at all. If not, we let 'em in.
		var count int
		err = a.DB.GetContext(ctx, &count, `SELECT count(*) FROM users`)
		if err != nil {
			return nil, err
		}

		// No users in DB, allow everyone
		if count == 0 {
			return &user, nil
		}

		return nil, auth.ErrAuthFailed
	}

	if err != nil {
		log.WithError(err).
			WithField("token", token).
			Error("got error during user lookup")
		return nil, err
	}

	return &user, nil
}

// Handler returns an HTTP handler that can be used to ask a user to authenticate themselves. It loads templates
// from the given FS.
func (a Provider) Handler(templateFS fs.FS) http.Handler {
	var (
		once sync.Once
		tpl  *template.Template
	)

	type responseData struct {
		Error string
	}

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		once.Do(func() {
			tpl = template.Must(template.ParseFS(templateFS, "templates/base.tpl", "templates/auth.tpl"))
		})

		if r.Method == "GET" {
			err := tpl.ExecuteTemplate(w, "auth.tpl", nil)
			if err != nil {
				log.WithError(err).
					Error("can't execute auth template")
			}

			return
		}

		// Get user and password from POST data
		userName := r.PostFormValue("username")
		password := r.PostFormValue("pass")

		log := log.WithField("user", userName)

		log.WithField("pass", password).
			Debug("attempting auth")

		// Get salted password from DB
		var userData struct {
			ID     int    `db:"id"`
			Salted string `db:"password"`
		}

		err := a.DB.GetContext(r.Context(), &userData,
			`SELECT id() AS id, password FROM users WHERE name = $1`,
			userName)

		if errors.Is(err, sql.ErrNoRows) {
			// Either no such user or no users at all
			var count int
			err = a.DB.GetContext(r.Context(), &count, `SELECT count(*) FROM users`)
			if err != nil {
				log.WithError(err).Error("can't get user count")
				http.Error(w, err.Error(), http.StatusInternalServerError)
				return
			}

			log.WithField("num_users", count).Info("got user count")
			// TODO: Deal with this. Report an error if != 0 users in db, otherwise allow them through.

			w.WriteHeader(http.StatusUnauthorized)

			rd := responseData{
				Error: "authentication failed",
			}

			log.WithField("data", rd).
				Error("writing response")

			err := tpl.ExecuteTemplate(w, "auth.tpl", rd)
			if err != nil {
				log.WithError(err).
					Error("can't execute auth template")
			}

			return
		}

		if err != nil {
			log.WithError(err).
				Error("can't get salted password")
		}

		hashed, err := hashPassword(r.Context(), a.DB, userName, password)
		if err != nil {
			log.WithError(err).
				Error("can't hash password")

			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}

		if hashed != userData.Salted {
			log.Error("password mismatch")

			w.WriteHeader(http.StatusUnauthorized)

			rd := responseData{
				Error: "authentication failed",
			}

			log.WithField("data", rd).
				Error("writing response")

			err := tpl.ExecuteTemplate(w, "auth.tpl", rd)
			if err != nil {
				log.WithError(err).
					Error("can't execute auth template")
			}

			return
		}

		log = log.WithField("user_id", userData.ID)

		// All good
		log.Info("auth looks good")

		addr := r.Header.Get("X-Forwarded-For")
		if addr == "" {
			addr = r.RemoteAddr
		}

		token, err := a.createSession(r.Context(), userData.ID, addr)
		if err != nil {
			log.WithError(err).
				Error("can't create session token")

			http.Error(w, err.Error(), http.StatusInternalServerError)

			return
		}

		cookie := http.Cookie{
			Name:     "session",
			Value:    token,
			Path:     "/",
			HttpOnly: true, // Might need to be unset if we do a REST-style API
		}

		log.WithField("cookie", cookie).
			Debug("setting token cookie")

		http.SetCookie(w, &cookie)

		// Redirect to initially requested URL
		http.Redirect(w, r, r.URL.String(), http.StatusFound)
	})
}

func (a Provider) createSession(ctx context.Context, userID int, remoteAddr string) (string, error) {
	var rawToken [40]byte

	n, err := crand.Read(rawToken[:])
	if err != nil {
		return "", err
	}

	if n != len(rawToken) {
		return "", errors.New("short read")
	}

	token := fmt.Sprintf("%02x", rawToken)

	tx, err := a.DB.Begin()
	if err != nil {
		return "", err
	}

	defer func() {
		if err != nil {
			tx.Rollback()
			return
		}

		tx.Commit()
	}()

	_, err = tx.ExecContext(ctx, `
		INSERT INTO sessions (user, token, remote, created)
		VALUES ($1, $2, $3, $4)`,
		userID, token, remoteAddr, time.Now())
	if err != nil {
		return "", err
	}

	return token, nil
}

func (a Provider) DeleteSession(ctx context.Context, token string) error {
	tx, err := a.DB.Begin()
	if err != nil {
		return err
	}

	defer func() {
		if err != nil {
			tx.Rollback()
			return
		}

		tx.Commit()
	}()

	_, err = tx.ExecContext(ctx, `DELETE FROM sessions WHERE token = $1`, token)
	if err != nil {
		return err
	}

	return nil
}

func (a Provider) ListSessions(ctx context.Context) ([]Info, error) {
	var sessions []Info

	err := a.DB.SelectContext(ctx, &sessions,
		`SELECT users.name AS name, sessions.created AS created, sessions.remote AS remote
		FROM users, sessions
		WHERE id(users) == sessions.user`)
	if err != nil {
		return nil, err
	}

	return sessions, nil
}

func (a Provider) ListUsers(ctx context.Context) ([]string, error) {
	var names []string

	err := a.DB.SelectContext(ctx, &names, `SELECT name FROM users`)
	return names, err
}

func (a Provider) UpdatePassword(ctx context.Context, userName, passOld, passNew string) (err error) {
	hashedOld, err := hashPassword(ctx, a.DB, userName, passOld)
	if err != nil {
		return err
	}

	hashedNew, err := hashPassword(ctx, a.DB, userName, passNew)
	if err != nil {
		return err
	}

	tx, err := a.DB.Beginx()
	if err != nil {
		return err
	}
	defer func() {
		if err != nil {
			tx.Rollback()
			return
		}

		tx.Commit()
	}()

	var pwHash []byte
	err = tx.GetContext(ctx, &pwHash, `SELECT password FROM users WHERE name = ?1`, userName)
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return err
	}

	if len(pwHash) != 0 && string(pwHash) != hashedOld {
		// A password exists already and it's not the one the user gave as their 'old' password.
		return auth.ErrAuthFailed
	}

	log.WithField("name", userName).
		Info("updating password")

	// Either user doesn't exist yet (no entry) or the password they gave as their old matched. We
	// can update.
	// Try an update first, if that fails, an insert
	res, err := tx.ExecContext(ctx, `UPDATE users SET password = ?1 WHERE name = ?2`, hashedNew, userName)
	if err != nil {
		return fmt.Errorf("updating password: %w", err)
	}

	n, err := res.RowsAffected()
	if err != nil {
		return err
	}

	if n == 0 {
		log.WithField("name", userName).
			Info("creating new user")

		_, err = tx.ExecContext(ctx, `INSERT INTO users (name, password) VALUES (?1, ?2)`, userName, hashedNew)
		if err != nil {
			return err
		}
	}

	return nil
}

func (a Provider) CreateUser(ctx context.Context, userName string) (string, error) {
	// generate random password
	const chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"

	var genPw string

	for idx := 0; idx < 16; idx++ {
		r := chars[rand.Intn(len(chars))]

		genPw += string(r)
	}

	err := a.UpdatePassword(ctx, userName, "", genPw)
	if err != nil {
		return "", fmt.Errorf("updating password for %q: %w", userName, err)
	}

	return genPw, nil
}

func (a Provider) DeleteUser(ctx context.Context, userName string) error {
	tx, err := a.DB.Beginx()
	if err != nil {
		return err
	}

	_, err = tx.ExecContext(ctx, `DELETE FROM users WHERE name = ?1`, userName)
	if err != nil {
		tx.Rollback()
		return err
	}

	return tx.Commit()
}
